""" base classes.  """

## imports

import _thread
import boht
import datetime
import importlib
import json
import logging
import os
import time

## exceptions

class ENOFILE(Exception):

    """ file doesnot exist. """

class ETYPE(Exception):

    """ wrong type. """

## classes

class Object:

    """ basic object with load/save capabilities """

    def __eq__(self, o):
        return self.__dict__ == o.__dict__

    def __ne__(self, o):
        return self.__dict__ != o.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __str__(self):
        """ return json string. """
        return json.dumps(self, default=default, indent=4, sort_keys=True)

    def get(self, k, d=None):
        """ getattr on this object. """
        return getattr(self, k, d)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def load(self, path):
        """ load json file into this object. """
        assert boht.cfg.storedir
        assert path
        logging.debug("load %s" % path)
        p = os.path.join(boht.cfg.storedir, path)
        if not os.path.exists(p):
            assert ENOFILE(path)
        with open(p) as f:
            d = json.load(f)
            self.update(d)
        return self

    def save(self, path="", stime=None, timed=False):
        """ save this object to disk. """
        assert boht.cfg.storedir
        if not path or stime or timed:
            if not stime:
                stime = str(datetime.datetime.now()).replace(" ", os.sep)
            path = os.path.join(boht.utils.get_type(self), stime)
        if os.sep in path:
            type = path.split(os.sep)[0]
        else:
            type = path
        if type not in boht.static.types:
            raise ETYPE(type)
        logging.info("save %s" % path)
        pp = os.path.join(boht.cfg.storedir, path)
        boht.utils.cdir(pp)
        with open(pp, "w") as f:
            json.dump(self, f, default=default, indent=4, sort_keys=True)
        return path

    def set(self, k, v):
        """ setattr on this object. """
        return setattr(self, k, v)

    def update(self, o, skip=False):
        """ update this object. """
        if "items" not in dir(o):
            o = vars(o)
        for k, v in o.items():
            if skip and not v:
                continue
            self.set(k, v)
        return self

class Store(Object):

    """ store objects on disk """

    def all(self, type, selector={}, index=None, delta=0, showdel=False):
        """ return all objects matching selector. """
        res = []
        nr = -1
        for fn in filenames(type, delta):
            nr += 1
            o = filehook(fn)
            if not showdel and ("_deleted" in dir(o) and o._deleted):
                continue
            if index is not None and nr != index:
                continue
            res.append((fn, o))
        return res

    def find(self, type, selector, index=None, delta=0, showdel=False):
        """ return specific objects matching selector. """
        nr = -1
        res = []
        for fn in filenames(type, delta):
            nr += 1
            o = filehook(fn)
            if not o:
                continue
            if not showdel and ("_deleted" in dir(o) and o._deleted):
                continue
            if search(o, selector):
                if index is not None and nr != index:
                    continue
                res.append((fn, o))
        return res

    def last(self, type, index=None, delta=0):
        """ return path, object of the last object of a type. """
        fns = filenames(type, delta)
        if fns:
            fn = fns[-1]
            return (fn, filehook(fn))
        return (None, None)

def default(o):
    """ convert an object to it's proper string representation. """
    t = type(o)
    if isinstance(o, Object):
        return vars(o)
    if isinstance(o, dict):
        return o.items()
    if isinstance(o, list):
        return iter(o)
    if t in [str, True, False, int, float]:
        return o
    return repr(o)

def filehook(fn):
    """ load an object from disk and convert it to it's proper type. """
    t = get_fntype(fn)
    o = get_cls(t)()
    o.load(fn)
    return o

def filenames(name, delta=0):
    """ fetch all filenames in a named subdir. """ 
    if not name:
        return []
    p = os.path.join(boht.cfg.storedir, name) + os.sep
    res = []
    now = time.time()
    past = now + delta
    for rootdir, dirs, files in os.walk(p, topdown=True):
        if name and name not in rootdir:
            continue
        for fn in files:
            fnn = os.path.join(rootdir, fn).split(boht.cfg.workdir)[-1][1:]
            if delta:
                if fn_time(fnn) < past:
                    continue
            res.append(os.sep.join(fnn.split(os.sep)[1:]))
    return sorted(res, key=lambda x: fn_time(x))

def fn_time(daystr):
    """ return timestamp derived from a string. """
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    try:
        datestr, rest = datestr.rsplit(".", 1)
    except ValueError:
        rest = ""
    try:
        t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
        if rest:
            t += float("." + rest)
    except ValueError:
        t = 0
    return t

def get_cls(name):
    """ convert a object name to its type. """
    modname, clsname = name.rsplit(".", 1)
    mod = importlib.import_module(modname)
    return getattr(mod, clsname)

def get_fntype(fn):
    """ retrieve type from filename. """
    return fn.split(os.sep)[0]

def hook(t, d):
    """ convert a dict to a typed object. """
    o = get_cls(t)()
    o.update(d)
    return o

def search(o, m: None):
    """ match given selector (dict with keys/values to match). """
    if not m:
        m = {}
    res = False
    for k, v in m.items():
        vv = o.get(k, None)
        if vv:
            if v == None:
                res = True
            elif v in str(vv):
                res = True
            else:
                res = False
                break
        else:
            res = False
            break
    return res
