""" event handler. """

## imports

import boht
import logging
import queue
import sys
import time
import threading
import types

## vars

handlers = {}

## classes

class Handler(boht.loader.Loader):

    """ basic event handler. """

    def __init__(self):
        super().__init__()
        self._queue = queue.Queue()
        self._ready = threading.Event()
        self._stopped = False
        self._threaded = True
        self._thrs = []
        self.state = boht.base.Object()
        self.verbose = False

    def dispatch(self, event):
        """ dispatch (proces) an event. """
        event.parse(event.txt)
        event._func = self.get_cmd(event)
        event.verbose = self.verbose
        if event._func:
            logging.warn("dispatch %s" % event.txt)
            try:
                event._func(event)
            except Exception as ex:
                logging.error(boht.trace.get_exception())
        event.show()
        event.ready()
        return event

    def get_cmd(self, event):
        """ return command if it exists. """
        return self.get_handler(event.cmd)

    def get_event(self):
        """ return event from queue (can be overloaded, but must block). """
        return self._queue.get()

    def get_handler(self, cmd):
        """ reutrn a handler matching a command. """
        val = handlers.get(cmd, None)
        if not val:
            modname = boht.static.names.get(cmd, None)
            if modname:
                self.load_mod(modname)
            val = handlers.get(cmd, None)
        return val

    def load_mod(self, name, mod=None):
        """ load a module. """
        mod = super().load_mod(name, mod)
        for key, func, modname in self.scan_commands(mod):
            handlers[key] = func
        return mod

    def loop(self):
        """ proces events. """
        while not self._stopped:
            try:
                event = self.get_event()
            except EOFError:
                break
            if not event:
                continue
            thr = boht.launch(self.dispatch, event)
            event._thrs.append(thr)

    def put(self, event):
        """ put an event on the queue. """
        self._queue.put_nowait(event)

    def ready(self):
        """ flag handler as ready. """
        self._ready.set()

    def register(self, cmd, handler):
        """ register a handler. """
        handlers[cmd] = handler

    def start(self):
        """ start the event loop. """
        boht.launch(self.loop)

    def stop(self):
        """ stop the handler. """
        self._stopped = True
        self._queue.put(None)

    def walk(self, modname, init=False):
        """ scan a module for commands/classes. """
        thrs = []
        mod = self.load_mod(modname)
        if init and "init" in dir(mod):
            thrs.append(boht.launch(mod.init))
        for m in dir(mod):
            module = getattr(mod, m)
            if type(module) != types.ModuleType:
                continue
            self.load_mod(module.__name__, module)
            if init and "init" in dir(module):
                thrs.append(boht.launch(module.init))
        return thrs

    def wait(self):
        """ wait for the handler. """
        return self._ready.wait()
