""" IRC bot module. """

## imports

import _thread
import boht
import logging
import os
import queue
import re
import socket
import ssl
import textwrap
import time
import threading

## initialisation

def init():
    bot = IRC()
    bot.start()
    return bot

## classes

class Cfg(boht.Cfg):

    """ irc config. """

    def __init__(self):
        super().__init__()
        self.blocking = 1
        self.channel = ""
        self.ipv6 = False
        self.nick = "boht"
        self.port = 6667
        self.realname = "boht"
        self.resume = ""
        self.server = ""
        self.ssl = False
        self.username = "boht"
        self.verbose = False

class IEvent(boht.Event):

    """ irc event. """

    def __init__(self):
        super().__init__()
        self.arguments = []
        self.cc = ""
        self.channel = ""
        self.command = ""
        self.nick = ""
        self.target = ""

class DEvent(boht.Event):

    """ dcc event. """

    def __init__(self):
        super().__init__()
        self._sock = None
        self._fsock = None
        self.channel = ""

class TextWrap(textwrap.TextWrapper):

    """ text wrapper (wrap around 500 chars lenght). """

    def __init__(self):
        super().__init__()
        self.drop_whitespace = False
        self.fix_sentence_endings = True
        self.replace_whitespace = True
        self.tabsize = 4
        self.width = 500

class IRC(boht.Bot):

    """ irc bot. """

    def __init__(self):
        super().__init__()
        self._buffer = []
        self._connected = threading.Event()
        self._outqueue = queue.Queue()
        self._sock = None
        self._fsock = None
        self._threaded = True
        self.cc = "!"
        self.cfg = Cfg()
        self.channels = []
        self.state = boht.base.Object()
        self.state.error = ""
        self.state.last = 0
        self.state.lastline = ""
        self.state.nrsend = 0
        self.state.pongcheck = False
        self.state.resume = None
        self.register("352", self.handler)
        self.register("366", self.handler)
        self.register("376", self.handle_connect)
        self.register("433", self.handler)
        self.register("ERROR", self.handle_error)
        self.register("MODE", self.handler)
        self.register("PING", self.handler)
        self.register("PONG", self.handler)
        self.register("PRIVMSG", self.handle_privmsg)

    def _connect(self):
        """ connect to the server. """
        if not self.cfg.server:
            logging.warn("need server, use -s option")
            return True
        if not self.cfg.channel:
            logging.warn("need channel, use -c option")
            return True
        oldsock = None
        if self.cfg.resume:
            b = boht.run.kernel.last("boht.irc.IRC")
            if b:
                fd = int(b.state["resume"])
                self.state.resume = fd
                logging.warn("resume %s" % fd)
            else:
                fd = self._sock.fileno()
            if self.cfg.ipv6:
                oldsock = socket.fromfd(fd , socket.AF_INET6, socket.SOCK_STREAM)
            else:
                oldsock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        if not oldsock:
            if self.cfg.ipv6:
                oldsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            else:
                oldsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        oldsock.setblocking(int(self.cfg.blocking or 0))
        oldsock.settimeout(60.0)
        if not self.cfg.resume:
            oldsock.connect((self.cfg.server, int(self.cfg.port or 6667)))
        oldsock.setblocking(int(self.cfg.blocking))
        oldsock.settimeout(700.0)
        if self.cfg.ssl:
            self._sock = ssl.wrap_socket(oldsock)
        else:
            self._sock = oldsock
        self.state.resume = self._sock.fileno()
        self._fsock = self._sock.makefile("r")
        os.set_inheritable(self.state.resume, os.O_RDWR)
        self._connected.set()
        logging.warning("connect %s %s" % (self.cfg.server, self.cfg.port or 6667))
        return True

    def _parsing(self, txt):
        """ parse incoming text into an irc event. """
        rawstr = str(txt)
        o = IEvent()
        o.txt = rawstr
        o.cc = self.cc
        o.command = ""
        o.arguments = []
        arguments = rawstr.split()
        o.origin = arguments[0]
        if o.origin.startswith(":"):
            o.origin = o.origin[1:]
            if len(arguments) > 1:
                o.command = arguments[1]
            if len(arguments) > 2:
                txtlist = []
                adding = False
                for arg in arguments[2:]:
                    if arg.startswith(":"):
                        adding = True
                        txtlist.append(arg[1:])
                        continue
                    if adding:
                        txtlist.append(arg)
                    else:
                        o.arguments.append(arg)
                    o.txt = " ".join(txtlist)
        else:
            o.cmd = o.command = o.origin
            o.origin = self.cfg.server
        try:
            o.nick, o.origin = o.origin.split("!")
        except ValueError:
            o.nick = ""
        if o.arguments:
            o.target = o.arguments[-1]
        else:
            o.target = ""
        if o.target.startswith("#"):
            o.channel = o.target
        else:
            o.channel = o.nick
        if not o.txt:
            if rawstr[0] == ":":
                rawstr = rawstr[1:]
            o.txt = rawstr.split(":", 1)[-1]
        if not o.txt and len(arguments) == 1:
            o.txt = arguments[1]
        o.args = o.txt.split()
        o.orig = repr(self)
        o.rest = " ".join(o.args)
        return o

    def _some(self, use_ssl=False, encoding="utf-8"):
        """ read from socket and put text into buffer. """
        if use_ssl:
            inbytes = self._sock.read()
        else:
            inbytes = self._sock.recv(512)
        txt = str(inbytes, encoding)
        if txt == "":
            raise ConnectionResetError
        txt = bytes(txt, "utf-8").decode("unicode_escape")
        self.state.lastline += txt
        splitted = self.state.lastline.split("\r\n")
        for s in splitted[:-1]:
            self._buffer.append(s)
            if not s.startswith("PING") and not s.startswith("PONG"):
                logging.warning(s.strip())
        self.state.lastline = splitted[-1]

    def announce(self, txt):
        """ announce on all channels. """
        for channel in self.channels:
            self.say(repr(self), channel, txt)

    def command(self, cmd, *args):
        """ send commands to the server. """
        if not args:
            self.raw(cmd)
            return
        if len(args) == 1:
            self.raw("%s %s" % (cmd.upper(), args[0]))
            return
        if len(args) == 2:
            self.raw("%s %s :%s" % (cmd.upper(), args[0], " ".join(args[1:])))
            return
        if len(args) >= 3:
            self.raw("%s %s %s :%s" % (cmd.upper(), args[0], args[1], " ".join(args[2:])))
            return

    def dispatch(self, event):
        """ dispatch on event.command (irc specific). """
        if event.command:
            func = self.get_handler(event.command)
            if func:
                func(event)
            return
        super().dispatch(event)

    def get_event(self):
        """ return event from buffer. """
        if not self._buffer:
            try:
                self._some()
            except socket.timeout as ex:
                e = IEvent()
                e._error = boht.trace.get_exception()
                e.command = "ERROR"
                return e
        return self._parsing(self._buffer.pop(0))

    def handle_connect(self, event):
        """ handler called on connect. """
        if "servermodes" in dir(self.cfg):
            self.raw("MODE %s %s" % (self.cfg.nick, self.cfg.servermodes))
        self.joinall()

    def handle_error(self, event):
        """ handler called on error. """
        logging.error(event.txt)
        self.state.error = event

    def handle_privmsg(self, event):
        """ handler called on privmsg. """
        boht.run.users.userhosts.set(event.nick, event.origin)
        if not boht.run.users.allowed(event.origin, "USER"):
            return
        if event.txt.startswith("\001DCC"):
            try:
                dcc = DCC()
                dcc.encoding = "utf-8"
                boht.launch(dcc.connect, event)
                return
            except ConnectionRefusedError:
                return
        elif event.txt.startswith("\001VERSION"):
            txt = "\001VERSION %s %s - %s\001" % ("BOHT", boht.__version__, boht.__txt__)
            self.command("NOTICE", event.channel, txt)
            return
        if event.txt and event.txt[0] == self.cc:
            txt = event.txt[1:]
            event.parse(txt)
            boht.run.kernel.put(event)

    def handler(self, event):
        """ handler for basic commands. """
        cmd = event.command
        if cmd == "PING":
            self.state.pongcheck = True
            self.command("PONG", event.txt)
        elif cmd == "PONG":
            self.state.pongcheck = False
        elif cmd == "451":
            nick = event.target + "_"
            self.cfg.nick = nick
            self.raw("USER %s" % self.cfg.username, True)
        elif cmd == "433":
            nick = event.target + "_"
            self.cfg.nick = nick
            self.raw("NICK %s" % self.cfg.nick, True)

    def joinall(self):
        """ join all channels. """
        for channel in self.channels:
            self.command("JOIN", channel)

    def logon(self, server, nick):
        """ send login credentials to the server. """
        self._connected.wait()
        self.raw("NICK %s" % nick, True)
        self.raw("USER %s %s %s :%s" % (self.cfg.username, server, server, self.cfg.realname), True)

    def loop(self):
        """ run the basic event loop. """
        try:
            self._connected.wait()
            super().loop()
        except ConnectionResetError:
            self.stop()
            self.start()

    def output(self):
        """ output thread. """
        self._connected.wait()
        while not self._stopped:
            channel, txt = self._outqueue.get()
            if not channel and not txt:
                break
            self.command("PRIVMSG", channel, txt)

    @boht.utils.locked
    def raw(self, txt, direct=False):
        """ send txt directly to the server. """
        self._connected.wait()
        txt = txt.rstrip()
        if self._stopped:
            return
        if not txt.startswith("PING") and not txt.startswith("PONG"):
            logging.warning(txt)
        if not txt.endswith("\r\n"):
            txt += "\r\n"
        txt = txt[:512]
        txt = bytes(txt, "utf-8")
        if not direct:
            if (time.time() - self.state.last) < 3.0:
                time.sleep(1.0 * (self.state.nrsend % 10))
        self.state.last = time.time()
        self.state.nrsend += 1
        self._sock.send(txt)

    def resume(self):
        """ resume the bot. """
        super().start()
        boht.launch(self.output)
        self.announce("done")

    def say(self, orig, channel, txt):
        """ say text on channel (orig is ignored). """
        wrapper = TextWrap()
        txt = str(txt)
        for line in txt.split("\n"):
            for t in wrapper.wrap(line):
                self._outqueue.put((channel, t))

    def start(self):
        """ start the bot. """
        p, last = boht.run.store.last("boht.irc.Cfg")
        if last:
            self.cfg.update(last)
        self.cfg.update(boht.cfg, skip=True)
        self.cfg.save(p)
        if not self.cfg.channel and boht.cfg.shell:
            logging.error("missing channel, use -c option")
            return
        if not self.cfg.server and boht.cfg.shell:
            logging.error("missing server, use -s option")
            return
        if self.cfg.channel:
            self.channels.append(self.cfg.channel)
        nr = 1
        self.state.error = ""
        super().start()
        boht.launch(self.output)
        boht.launch(self.connect)

    def connect(self):
        """ loop until connected. """
        nr = 0
        while 1:
            try:
                self._connect()
                break
            except socket.gaierror as ex:
                logging.warn(ex)
                break
            except ConnectionRefusedError:
                logging.warn("%s connection refused" % self.cfg.server)
            time.sleep(nr * 3.0)
            nr += 1
        if not self.cfg.resume and not self.state.error:
            self.logon(self.cfg.server, self.cfg.nick)

    def stop(self):
        """ stop the bot. """
        self._stopped = True
        super().stop()
        self._outqueue.put_nowait((None, None))
        self.ready()

class DCC(boht.Bot):

    """ direct client to client (bot) chat. """

    def __init__(self):
        super().__init__()
        self._sock = None
        self._fsock = None
        self.encoding = "utf-8"
        self.origin = ""

    def announce(self, txt):
        """ echo text to the client. """
        self.raw(txt)

    def connect(self, event):
        """ connect to the clients irc client. """
        arguments = event.txt.split()
        addr = arguments[3]
        port = arguments[4][:-1]
        port = int(port)
        if re.search(':', addr):
            s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        else:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((addr, port))
        s.send(bytes('Welcome to %s %s !!\n' % ("BOHT", event.nick), "utf-8"))
        s.setblocking(True)
        os.set_inheritable(s.fileno(), os.O_RDWR)
        self._sock = s
        self._fsock = self._sock.makefile("rw")
        self.origin = event.origin
        super().start()

    def errored(self, event):
        """ handler called on error. """
        self.state.error = event
        logging.error(str(event))

    def event(self, txt):
        """ create a dcc event. """
        e = DEvent()
        e.txt = txt
        e._sock = self._sock
        e._fsock = self._fsock
        e.orig = repr(self)
        e.channel = self.origin
        e.origin = self.origin or "root@dcc"
        return e

    def get_event(self):
        """ return a event (block on client input). """
        txt = self._fsock.readline()
        return self.event(txt)

    def raw(self, txt):
        """ output raw text on the socket. """
        self._fsock.write(txt.rstrip())
        self._fsock.write("\n")
        self._fsock.flush()

    def say(self, orig, channel, txt):
        """ echo text to client. """
        self.raw(txt)
