""" utility functions. """

## imports

import _thread
import html
import html.parser
import importlib
import json
import logging
import os
import re
import sys
import time
import threading
import traceback
import types
import urllib
import urllib.error
import urllib.parse
import urllib.request

from random import randint
from datetime import datetime

## defines

timestrings = [
    "%a, %d %b %Y %H:%M:%S %z",
    "%d %b %Y %H:%M:%S %z",
    "%d %b %Y %H:%M:%S",
    "%a, %d %b %Y %H:%M:%S",
    "%d %b %a %H:%M:%S %Y %Z",
    "%d %b %a %H:%M:%S %Y %z",
    "%a %d %b %H:%M:%S %Y %z",
    "%a %b %d %H:%M:%S %Y",
    "%d %b %Y %H:%M:%S",
    "%a %b %d %H:%M:%S %Y",
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%dt%H:%M:%S+00:00",
    "%a, %d %b %Y %H:%M:%S +0000",
    "%d %b %Y %H:%M:%S +0000",
]

year_formats = [
    "%b %H:%M",
    "%b %H:%M:%S",
    "%a %H:%M %Y",
    "%a %H:%M",
    "%a %H:%M:%S",
    "%Y-%m-%d",
    "%d-%m-%Y",
    "%d-%m",
    "%m-%d",
    "%Y-%m-%d %H:%M:%S",
    "%d-%m-%Y %H:%M:%S",
    "%d-%m %H:%M:%S",
    "%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M",
    "%d-%m-%Y %H:%M",
    "%d-%m %H:%M",
    "%m-%d %H:%M",
    "%H:%M:%S",
    "%H:%M"
]

## functions

def cdir(path):
    """ create a directory. """
    if os.path.exists(path):
        return
    res = ""
    path = os.path.dirname(path)
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        padje = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(padje)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
        except OSError as ex:
            logging.error(get_exception())
    return True

def day():
    """ return timestamp of today 00:00 hours. """
    return str(datetime.today()).split()[0]

def days(path):
    """ calculate days since object creation. """
    from boht.base import fn_time
    return elapsed(time.time() - fn_time(path))

def elapsed(seconds, short=True):
    """ return elapsed time in string form. """
    txt = ""
    nsec = float(seconds)
    year = 365*24*60*60
    week = 7*24*60*60
    nday = 24*60*60
    hour = 60*60
    minute = 60
    years = int(nsec/year)
    nsec -= years*year
    weeks = int(nsec/week)
    nsec -= weeks*week
    nrdays = int(nsec/nday)
    nsec -= nrdays*nday
    hours = int(nsec/hour)
    nsec -= hours*hour
    minutes = int(nsec/minute)
    sec = nsec - minutes*minute
    if years:
        txt += "%sy" % years
    if weeks:
        nrdays += weeks * 7
    if nrdays:
        txt += "%sd" % nrdays
    if years and short and txt:
        return txt
    if hours:
        txt += "%sh" % hours
    if nrdays and short and txt:
        return txt
    if minutes:
        txt += "%sm" % minutes
    if hours and short and txt:
        return txt
    if sec == 0:
        txt += "0s"
    elif sec < 1 or not short:
        txt += "%.3fs" % sec
    else:
        txt += "%ss" % int(sec)
    txt = txt.strip()
    return txt

def format(o, keys=None, full=False):
    """ format an object into a displayable string. """
    if keys is None:
        keys = vars(o).keys()
    res = []
    txt = ""
    for k in keys:
        v = getattr(o, k, None)
        if not v:
            continue
        v = str(v)
        if full:
            res.append("%s=%s " % (k, v))
        else:
            res.append(v)
    for v in res:
        txt += "%s " % v.strip()
    return txt.strip()

def get_cls(name):
    """ convert a object name to its type. """
    modname, clsname = name.rsplit(".", 1)
    mod = importlib.import_module(modname)
    return getattr(mod, clsname)

def get_type(o):
    """ determine type of a object (module/class name). """
    mod = ""
    try:
        mod = o.__self__.__class__.__module__
    except AttributeError:
        mod = o.__class__.__module__
    nn = ""
    try:
        nn = o.__self__.__class__.__name__
    except AttributeError:
        nn = o.__class__.__name__
    return "%s.%s" % (mod, nn)

def get_name(o):
    """ return module.classname of an object. """
    t = type(o)
    if t == types.ModuleType:
        return o.__name__
    try:
        n = "%s.%s" % (o.__self__.__class__.__name__, o.__name__)
    except AttributeError:
        try:
            n = "%s.%s" % (o.__class__.__name__, o.__name__)
        except AttributeError:
            try:
                n = o.__class__.__name__
            except AttributeError:
                n = o.__name__
    return n

def get_hour(daystr):
    """ parse string for hour notation. """
    try:
        hmsre = re.search(r'(\d+):(\d+):(\d+)', str(daystr))
        hours = 60 * 60 * (int(hmsre.group(1)))
        hoursmin = hours  + int(hmsre.group(2)) * 60
        hms = hoursmin + int(hmsre.group(3))
    except AttributeError:
        pass
    except ValueError:
        pass
    try:
        hmre = re.search(r'(\d+):(\d+)', str(daystr))
        hours = 60 * 60 * (int(hmre.group(1)))
        hms = hours + int(hmre.group(2)) * 60
    except AttributeError:
        return 0
    except ValueError:
        return 0
    return hms

def get_time(daystr):
    """ return timestamp. """
    for f in year_formats:
        try:
            t = time.mktime(time.strptime(daystr, f))
            return t
        except Exception:
            pass

def get_url(*args):
    """ perform a GET on a url. """
    url = urllib.parse.urlunparse(urllib.parse.urlparse(args[0]))
    req = urllib.request.Request(url, headers={"User-Agent": useragent()})
    resp = urllib.request.urlopen(req)
    resp.data = resp.read()
    logging.debug("GET %s %s %s" % (url, resp.status, resp.reason))
    return resp

def locked(func):

    """ basic locker decorator. """

    lock = _thread.allocate_lock()

    def lockedfunc(*args, **kwargs):
        """ locked function %s """ % get_name(func)
        lock.acquire()
        res = None
        try:
            res = func(*args, **kwargs)
        finally:
            try:
                lock.release()
            except Exception:
                pass
        return res

    return lockedfunc

def now():
    """ return current timestamp. """
    return str(datetime.now()).split()[0]

def parse_date(daystr):
    """ parse a timestring into a timestamp. """
    sec = 0
    min = 0
    hour = 0
    day = 0
    w = 0
    diff = 0
    prev = ""
    val = 0
    total = 0
    for c in daystr:
        if c not in ["s", "m", "h", "d", "w", "y"]:
            val += c
            continue
        if c == "y":
            total += val * 3600*24*365
        if c == "w":
            total += val * 3600*24*7
        elif c == "d":
            total += val * 3600*24
        elif c == "h":
            total += val * 3600
        elif c == "m":
            total += val * 60
        else:
            total += val
        val = 0
    return total

def rtime():
    """ return a day/time string to use for saving. """
    import datetime
    res = str(datetime.datetime.now()).replace(" ", os.sep)
    return res

def sliced(o, keys=None):
    """return a sliced object. """
    oo = boht.base.Object()
    if not keys:
        keys = o.keys()
    for k in keys:
        try:
            oo[k] = o[k]
        except KeyError:
            pass
    return oo

def strip_html(text):
    """ strip html characters from string. """
    clean = re.compile('<.*?>')
    return re.sub(clean, '', text)

def stripped(jid):
    """ strip resource from jid. """
    try:
        return str(jid).split("/")[0]
    except (IndexError, ValueError):
        return str(jid)

def to_day(daystring):
    """ convert date/time to timestamp. """
    line = ""
    daystr = str(daystring)
    for word in daystr.split():
        if "-" in word:
            line += word + " "
        elif ":" in word:
            line += word
    if "-" not in line:
        line = day() + " " + line
    try:
        return get_time(line.strip())
    except ValueError:
        pass

def to_time(daystr):
    """ convert time to timestamp. """
    daystr = daystr.strip()
    if "," in daystr:
        daystr = " ".join(daystr.split(None)[1:7])
    elif "(" in daystr:
        daystr = " ".join(daystr.split(None)[:-1])
    else:
        try:
            d, h = daystr.split("T")
            h = h[:7]
            daystr = " ".join([d, h])
        except (ValueError, IndexError):
            pass
    res = 0
    for tstring in timestrings:
        try:
            res = time.mktime(time.strptime(daystr, tstring))
        except ValueError:
            try:
                res = time.mktime(time.strptime(" ".join(daystr.split()[:-1]), tstring))
            except ValueError:
                pass
        if res:
            break
    return res

def useragent():
    """ return useragent. """
    return 'Mozilla/5.0 (X11; Linux x86_64) BOHT +https://bitbucket.org/bthate/boht)'

def unescape(text):
    """ unescape control characters. """
    txt = re.sub(r"\s+", " ", text)
    return html.parser.HTMLParser().unescape(txt)
